/*!
  \file
  \brief LMS ログの読みだし

  \author Satofumi Kamimura
*/

#include <cstdlib>
#include <cmath>
#include <sstream>
#include <fstream>
#include "Lms_log_reader.h"
#include "Lms_driver.h"
#include "token_utils.h"

using namespace hrk;
using namespace std;


namespace
{
    enum {
        First_data_packet = 2,
    };

    const double Default_scan_usec = 20000.0;

    typedef struct
    {
        fstream::pos_type seek_position;
        long total_play_second;
        long msec_to_next_scan;
    } scan_t;
}


struct Lms_log_reader::pImpl
{
    string error_message_;
    string log_file_;
    ifstream* fin_;
    bool is_loaded_;
    vector<scan_t> scan_data_;
    bool with_intensity_;
    bool is_multiecho_;

    int start_index_;
    int stop_index_;
    int skip_steps_;
    int scan_times_;
    int scan_skips_;


    pImpl()
        : error_message_("no error."), fin_(NULL), is_loaded_(false),
          with_intensity_(false), is_multiecho_(false),
          start_index_(0), stop_index_(0), skip_steps_(1),
          scan_times_(0), scan_skips_(0)
    {
    }


    ~pImpl()
    {
        close();
    }


    bool load(const char* log_file)
    {
        close();

        log_file_ = log_file;
        if (!reset_fin()) {
            error_message_ = "open fail.";
            return false;
        }

        // データ量の読みだし
        enum { Invalid_timestamp = -1 };
        //long last_timestamp = Invalid_timestamp;
        double total_msec = 0.0;
        double scan_usec = Default_scan_usec;
        fstream::pos_type last_position = 0;
        //bool last_LF_detected = false;
        //bool command_parsed = false;
        //size_t count_in_command = 0;
        bool is_config_loaded_ = false;
        string line;
        long packet_number = 0;

        const char ETX = '\03';
        while (getline(*fin_, line, ETX)) {
            ++packet_number;

            // 読み出し中のデータフォーマットが適切かを確認する
            // !!!

            const string command_token = lms_command_token(line);

            if (command_token == "sRA LMPscancfg") {
                // 最初のパケットは計測の設定
                with_intensity_ = false;
                is_multiecho_ = false;

                // !!! 設定のパケットから読み出すようにする
                scan_skips_ = 0;
                scan_times_ = 0;
                start_index_ = 0;
                stop_index_ = 380;
                skip_steps_ = 0;
                scan_usec = 20000;
                continue;

            } else if ((command_token == "sEA LMDscandata") ||
                       (command_token == "sRA LMPscancfg")) {
                // 計測開始, 情報の取得のコマンドは無視する
                continue;

            } else if (command_token == "sRA DeviceIdent") {
                // DeviceIdent コマンドの受信で設定の読み出しが終わったとみなす
                is_config_loaded_ = true;
                continue;
            }

            if (!is_config_loaded_) {
                continue;
            }

            // 距離データを登録する
            last_position = fin_->tellg();
            long msec_to_next_scan = scan_usec / 1000.0;
            total_msec += scan_usec / 1000.0;

            scan_t scan;
            scan.seek_position = last_position;
            scan.msec_to_next_scan = msec_to_next_scan;
            scan.total_play_second = ceil(total_msec / 1000.0);
            scan_data_.push_back(scan);
        }
        is_loaded_ = scan_data_.empty() ? false : true;
        if (!is_loaded_) {
            error_message_ = "no scan data.";
            return false;
        }

        scan_t &last = scan_data_.back();
        last.total_play_second = ceil(total_msec / 1000.0);
        last.msec_to_next_scan = 0;

        if (!reset_fin()) {
            error_message_ = "reopen fail.";
            return false;
        }
        return true;
    }


    string lms_command_token(const string& line)
    {
        const vector<string> tokens = split(line, " ");
        if (tokens.size() < 2) {
            return "";
        }

        return tokens[0].substr(1) + ' ' + tokens[1];
    }


    void close()
    {
        delete fin_;
        fin_ = NULL;

        is_loaded_ = false;
        scan_data_.clear();
    }


    bool reset_fin()
    {
        delete fin_;
        fin_ = new ifstream(log_file_.c_str(), ios_base::binary);
        return fin_->is_open();
    }
};


Lms_log_reader::Lms_log_reader() : pimpl(new pImpl)
{
}


Lms_log_reader::~Lms_log_reader()
{
}


bool Lms_log_reader::load(const char* log_file)
{
    return pimpl->load(log_file);
}


void Lms_log_reader::log_measurement_type(bool& with_intensity,
                                          bool& is_multiecho)
{
    with_intensity = pimpl->with_intensity_;
    is_multiecho = pimpl->is_multiecho_;
}


void Lms_log_reader::log_range(int& first_step, int& last_step, int& skip_steps)
{
    first_step = pimpl->start_index_;
    last_step = pimpl->stop_index_;
    skip_steps = pimpl->skip_steps_;
}


void Lms_log_reader::log_scans(int& scan_times, int& scan_skips)
{
    scan_skips = pimpl->scan_skips_;
    scan_times = pimpl->scan_times_;
}


size_t Lms_log_reader::total_sec() const
{
    if (!pimpl->is_loaded_) {
        return 0;
    }

    if (pimpl->scan_data_.empty()) {
        return 0;
    }

    return pimpl->scan_data_.back().total_play_second;
}


size_t Lms_log_reader::total_scans() const
{
    if (!pimpl->is_loaded_) {
        return 0;
    }

    return pimpl->scan_data_.size();
}


bool Lms_log_reader::scan_data(size_t scan_index,
                               long& total_play_second, long& msec_to_next_scan)
{
    if (!pimpl->is_loaded_) {
        return false;
    }

    if (pimpl->scan_data_.empty()) {
        return 0;
    }

    scan_t& scan = pimpl->scan_data_[scan_index];
    total_play_second = scan.total_play_second;
    msec_to_next_scan = scan.msec_to_next_scan;

    return true;
}


bool Lms_log_reader::reload()
{
    if (!pimpl->is_loaded_) {
        return false;
    }

    return pimpl->reset_fin();
}


bool Lms_log_reader::set_next_scans(int scan_index)
{
    if (!pimpl->is_loaded_) {
        return false;
    }

    if ((scan_index < 0) ||
        (scan_index >= static_cast<int>(pimpl->scan_data_.size()))) {
        return false;
    }

    scan_t& scan = pimpl->scan_data_[scan_index];
    pimpl->fin_->seekg(scan.seek_position);
    return true;
}


const char* Lms_log_reader::what() const
{
    return pimpl->error_message_.c_str();
}


bool Lms_log_reader::change_baudrate(long baudrate)
{
    static_cast<void>(baudrate);

    // 何もしない

    return true;
}


bool Lms_log_reader::is_open() const
{
    return pimpl->fin_->eof() ? false : true;
}


void Lms_log_reader::close()
{
    pimpl->close();
}


int Lms_log_reader::write(const char* data, size_t data_size)
{
    static_cast<void>(data);
    static_cast<void>(data_size);

    // 正常に書き込めたことにする
    // 特に何もしない

    return data_size;
}


int Lms_log_reader::read(char* data, size_t max_data_size, int timeout)
{
    static_cast<void>(timeout);

    if (pimpl->fin_->eof()) {
        return -1;
    }

    pimpl->fin_->read(data, max_data_size);
    return pimpl->fin_->gcount();
}


void Lms_log_reader::ungetc(int ch)
{
    pimpl->fin_->putback(ch);
}
